@@ -17,11 +17,10 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
- * 02111-1307  USA
+ * along with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  **********************************************************************/
 
@@ -60,7 +59,7 @@ int dbserver_get(
   }
   if ((ret = dbp->open(
        dbp,
-#ifdef HAVE_BDB_4_1
+#ifdef HAVE_BDB_OPEN_7
        NULL,
 #endif
        (options_str==NULL)?PERDITIONDB_BDB_DEFAULT_MAPNAME:options_str,
