@@ -425,7 +425,7 @@ void io_set_err(io_t *io, enum io_err err);
 
 
 /**********************************************************************
- * io_get_timoeut
+ * io_get_timeout
  * Get the idle timeout of an io
  * pre: io: io_t to get the idle timeout of
  * post: none
@@ -439,7 +439,7 @@ long io_get_timeout(io_t *io)
 
 
 /**********************************************************************
- * io_set_timoeut
+ * io_set_timeout
  * Set the idle timeout of an io
  * pre: io: io_t to set the idle timeout of
  *      timeout: timeout in seconds, 0 for no timeout
@@ -655,7 +655,7 @@ err:
  *         0 otherwise (one of io_a or io_b closes gracefully)
  **********************************************************************/
 
-static int __io_pipe_read(int fd, void *buf, size_t count, void *data){
+static ssize_t __io_pipe_read(int fd, void *buf, size_t count, void *data){
   io_t *io;
   io_select_t *s;
   ssize_t bytes;
@@ -693,7 +693,9 @@ err:
 }
          
 
-static int __io_pipe_write(int fd, const void *buf, size_t count, void *data){
+static ssize_t
+__io_pipe_write(int fd, const void *buf, size_t count, void *data)
+{
   io_t *io;
   io_select_t *s;
 
@@ -709,9 +711,9 @@ static int __io_pipe_write(int fd, const void *buf, size_t count, void *data){
 }
 
          
-ssize_t io_pipe(io_t *io_a, io_t *io_b, unsigned char *buffer,
-      int buffer_length, int *return_a_read_bytes,
-      int *return_b_read_bytes, timed_log_t *log){
+ssize_t io_pipe(io_t *io_a, io_t *io_b, char *buffer,
+      int buffer_length, size_t *return_a_read_bytes,
+      size_t *return_b_read_bytes, timed_log_t *log){
   int bytes;
   io_select_t *s;
   long timeout;
@@ -748,13 +750,13 @@ ssize_t io_pipe(io_t *io_a, io_t *io_b, unsigned char *buffer,
   if((bytes=vanessa_socket_pipe_func(io_get_rfd(io_a), io_get_wfd(io_a),
       io_get_rfd(io_b), io_get_wfd(io_b), buffer, buffer_length,
       timeout, return_a_read_bytes, return_b_read_bytes,
-      __io_pipe_read, __io_pipe_write, io_select, (void *)s))<0){
-	VANESSA_LOGGER_DEBUG("vanessa_socket_pipe_func");
+      __io_pipe_read, __io_pipe_write, io_select, (void *)s))) {
         if (bytes == 1) {
             /* timeout */
             io_a->err = io_err_timeout;
             io_b->err = io_err_timeout;
-        }
+        } else
+            VANESSA_LOGGER_DEBUG("vanessa_socket_pipe_func");
         goto err;
   }
 
@@ -763,9 +765,9 @@ ssize_t io_pipe(io_t *io_a, io_t *io_b, unsigned char *buffer,
   return(bytes);
 
 err:
-  if (io_a->err != io_err_none && io_b->err != io_err_none) {
+  if (io_a->err == io_err_none)
             io_a->err = io_err_other;
+  if (io_b->err == io_err_none)
             io_b->err = io_err_other;
-  }
   return -1;
 }
