@@ -94,6 +94,9 @@ struct sockaddr_storage *sockname;
 /* PID file that has been created */
 static char *pid_file;
 
+/* Programme name */
+static char *progname;
+
 /*
  * Used for opening dynamic server lookup library
  * Kept global so they can be used in signal handlers
@@ -114,28 +117,8 @@ static int perdition_chown(const char *path, const char *username,
 
 /* Macro to clean things up when we jump around in the main loop*/
 #define PERDITION_CLEAN_UP_MAIN \
-  if(username!=NULL && username!=pw.pw_name && username!=pw2.pw_name){ \
-    free(username); \
-  } \
-  username=NULL; \
-  if(pw2.pw_name!=NULL && pw2.pw_name!=pw.pw_name){ \
-    free(pw2.pw_name); \
-  } \
-  pw2.pw_name=NULL; \
-  pw2.pw_passwd=NULL; \
-  if(pw.pw_name!=NULL){ \
-    free(pw.pw_name); \
-    pw.pw_name=NULL; \
-  } \
-  if(pw.pw_passwd!=NULL){ \
-    free(pw.pw_passwd); \
-    pw.pw_passwd=NULL;\
-  } \
-  if(pw2.pw_name!=NULL){ \
-    free(pw2.pw_name); \
-    pw2.pw_name=NULL; \
-  } \
-  pw2.pw_passwd=NULL; \
+  auth_free_data(&auth); \
+  auth_zero(&auth); \
   if (!round_robin_server){ \
     user_server_port_destroy(usp); \
   } \
@@ -148,9 +131,7 @@ static int perdition_chown(const char *path, const char *username,
   server_io=NULL; \
   usp=NULL; \
   token_destroy(&client_tag); \
-  tls_state=0; \
-  opt.capability = protocol->capability(opt.capability, \
-    &(opt.mangled_capability), opt.ssl_mode, tls_state);
+  tls_state=0;
 
 /* Macro to set the uid and gid */
 #define PERDITION_SET_UID_AND_GID \
@@ -159,130 +140,242 @@ static int perdition_chown(const char *path, const char *username,
     perdition_exit_cleanly(-1); \
   }
 
+struct quoted_str {
+	const char *quote, *str;
+};
+
+static struct quoted_str quote_str(const char *in)
+{
+	struct quoted_str out;
+
+	if (in) {
+		out.quote = "\"";
+		out.str = in;
+	} else {
+		out.quote = "";
+		out.str = "NONE";
+	}
+
+	return out;
+}
+
 static void 
 perdition_log_auth(timed_log_t *auth_log, const char *from_to_host_str,
-		struct passwd *pw, const char *servername, const char *port, 
-		const char *progname, const char *reason)
+		struct auth *auth, const char *servername, const char *port,
+		const char *reason, int ssl_mode, int tls_state)
 {
-	char *passwd;
+	const char *eu_ssl, *rs_ssl, *pw_head, *pw_body, *pw_tail;
+	struct quoted_str authorisation_id = quote_str(auth->authorisation_id);
+	char *protocol = NULL;
 
-	if (!strcmp(reason, "ok")) {
-		if (opt.log_passwd & LOG_PASSWD_OK)
-			passwd = pw->pw_passwd;
-		else 
-			passwd = "XXXXXX";
-	}
-	else {
-		if (opt.log_passwd & LOG_PASSWD_FAIL)
-			passwd = pw->pw_passwd;
-		else 
-			passwd = "XXXXXX";
+	if (opt.debug &&
+	    ((!strcmp(reason, "ok") && (opt.log_passwd & LOG_PASSWD_OK)) ||
+	     (opt.log_passwd & LOG_PASSWD_FAIL))) {
+		pw_body = auth->passwd;
+		pw_head = " passwd=\"";
+		pw_tail = "\"";
 	}
+	else
+		pw_head = pw_body = pw_tail = "";
+
+	if (ssl_mode & SSL_MODE_SSL_LISTEN)
+		eu_ssl = "ssl";
+	else if (tls_state & SSL_MODE_TLS_LISTEN)
+		eu_ssl = "starttls";
+	else
+		eu_ssl = "plaintext";
+
+	if (ssl_mode & SSL_MODE_SSL_OUTGOING)
+		rs_ssl = "ssl";
+	else if (tls_state & SSL_MODE_TLS_OUTGOING)
+		rs_ssl = "starttls";
+	else
+		rs_ssl = "plaintext";
+
+	protocol = protocol_list(protocol, NULL, opt.protocol);
 
 	memset(auth_log->log_str, 0, sizeof(auth_log->log_str));
 	snprintf(auth_log->log_str, sizeof(auth_log->log_str),
-			"Auth:%s user=\"%s\" passwd=\"%s\" "
-			"server=\"%s\" port=\"%s\" status=\"%s\"",
-			from_to_host_str, str_null_safe(pw->pw_name),
-			str_null_safe(passwd), str_null_safe(servername),
-			str_null_safe(port), str_null_safe(reason));
+			"Auth:%s client-secure=%s authorisation_id=%s%s%s "
+			"authentication_id=\"%s\"%s%s%s "
+			"server=\"%s:%s\" protocol=%s "
+			"server-secure=%s status=\"%s\"",
+			from_to_host_str, eu_ssl, authorisation_id.quote,
+			authorisation_id.str, authorisation_id.quote,
+			str_null_safe(auth->authentication_id),
+			pw_head, pw_body, pw_tail, str_null_safe(servername),
+			str_null_safe(port), str_null_safe(protocol),
+			rs_ssl, str_null_safe(reason));
 	auth_log->log_time = time(NULL) + opt.connect_relog;
 
+	free(protocol);
+
 	VANESSA_LOGGER_LOG(LOG_NOTICE, auth_log->log_str);
 
-	set_proc_title("%s: auth %s", progname, str_null_safe(reason));
+	set_proc_title("%s: connect (%s)", progname,
+		       str_null_safe(auth_get_authorisation_id(auth)));
 }
 
 #define PERDITION_LOG_AUTH(_reason)                                        \
 do {                                                                       \
-	perdition_log_auth(&auth_log, from_to_host_str, &pw, servername,   \
-			port, progname, _reason);                          \
+	perdition_log_auth(&auth_log, from_to_host_str, &auth, servername, \
+			port, _reason, opt.ssl_mode, tls_state);           \
 } while(0)
 
 static void 
 login_failed_protocol(protocol_t *protocol, int protocol_type, 
 		io_t *io, token_t *tag, timed_log_t *auth_log, 
-		const char *from_to_host_str, struct passwd *pw,
+		const char *from_to_host_str, struct auth *auth,
 		const char *servername, const char *port,
-		const char *progname, const char *reason)
+		const char *reason, int ssl_mode, int tls_state)
 {
 	sleep(PERDITION_AUTH_FAIL_SLEEP);
-	if (protocol->write(io, NULL_FLAG, tag, protocol->type[protocol_type], 
-				0, reason) < 0) {
+	if (protocol->write_str(io, NULL_FLAG, tag,
+				protocol->type[protocol_type], reason) < 0) {
 		VANESSA_LOGGER_DEBUG("protocol->write");
 		VANESSA_LOGGER_ERR("Fatal error writing to client. "
 				"Exiting child.");
 		perdition_exit_cleanly(-1);
 	}
 
-	perdition_log_auth(auth_log, from_to_host_str, pw, servername, port,
-			progname, reason);
+	perdition_log_auth(auth_log, from_to_host_str, auth, servername, port,
+			   reason, ssl_mode, tls_state);
+}
+
+void logger_init(void)
+{
+	vanessa_logger_t *vl;
+
+	vl = vanessa_logger_openlog_filehandle(stderr, progname, LOG_DEBUG,
+		       VANESSA_LOGGER_F_CONS|VANESSA_LOGGER_F_NO_IDENT_PID);
+	if (!vl) {
+		fprintf(stderr, "main: vanessa_logger_openlog_syslog\n"
+			"Fatal error opening logger. Exiting.\n");
+		perdition_exit_cleanly(-1);
+	}
+
+	vanessa_logger_set(vl);
+}
+
+void logger_reopen(FILE *fh)
+{
+	vanessa_logger_t *vl;
+
+	if (fh != NULL)
+		vl = vanessa_logger_openlog_filehandle(fh, progname,
+			opt.debug?LOG_DEBUG:(opt.quiet?LOG_ERR:LOG_INFO),
+			VANESSA_LOGGER_F_CONS|VANESSA_LOGGER_F_TIMESTAMP);
+	else if (opt.log_facility != NULL && *(opt.log_facility) == '/')
+		vl = vanessa_logger_openlog_filename(opt.log_facility,
+			progname,
+			opt.debug?LOG_DEBUG:(opt.quiet?LOG_ERR:LOG_INFO),
+			VANESSA_LOGGER_F_CONS|VANESSA_LOGGER_F_TIMESTAMP);
+	else
+		vl = vanessa_logger_openlog_syslog_byname(opt.log_facility,
+			progname,
+			opt.debug?LOG_DEBUG:(opt.quiet?LOG_ERR:LOG_INFO),
+			LOG_CONS);
+
+	if (vl == NULL) {
+    		VANESSA_LOGGER_DEBUG("vanessa_logger_openlog");
+		VANESSA_LOGGER_ERR("Fatal error opening logger. Exiting.");
+		perdition_exit_cleanly(-1);
+	}
+
+	vanessa_logger_closelog(vanessa_logger_get());
+	vanessa_logger_set(vl);
 }
 
 #define LOGIN_FAILED_PROTOCOL(_type, _reason)                               \
 do {                                                                        \
 	login_failed_protocol(protocol, _type, client_io, client_tag,       \
-			&auth_log, from_to_host_str, &pw, servername,       \
-			port, progname, "failed: " _reason);                \
+			&auth_log, from_to_host_str, &auth, servername,     \
+			port, "failed: " _reason, opt.ssl_mode, tls_state); \
 } while(0)
 
+static void perdition_log_close_early(const char *from_to_host_str,
+				      struct auth *auth)
+{
+	struct quoted_str authorisation_id = quote_str(auth->authorisation_id);
+
+	VANESSA_LOGGER_INFO_UNSAFE("Closing NULL session:%s "
+				   "authorisation_id=%s%s%s "
+				   "authentication_id=\"%s\"",
+				   from_to_host_str,
+				   authorisation_id.quote,
+				   authorisation_id.str,
+				   authorisation_id.quote,
+				   str_null_safe(auth->authentication_id));
+}
+
+static void perdition_log_close(const char *from_to_host_str,
+				struct auth *auth, size_t received, size_t sent)
+{
+	struct quoted_str authorisation_id = quote_str(auth->authorisation_id);
+
+	VANESSA_LOGGER_INFO_UNSAFE("Closing session:%s "
+				   "authorisation_id=%s%s%s "
+				   "authentication_id=\"%s\" "
+				   "received=%zu sent=%zu",
+				   from_to_host_str,
+				   authorisation_id.quote,
+				   authorisation_id.str,
+				   authorisation_id.quote,
+				   str_null_safe(auth->authentication_id),
+				   received, sent);
+}
+
 /**********************************************************************
  * Muriel the main function
  **********************************************************************/
 
 int main (int argc, char **argv, char **envp){
-  vanessa_logger_t *vl;
-  struct passwd pw = {NULL, NULL};
-  struct passwd pw2 = {NULL, NULL};
-  unsigned char *server_resp_buf=NULL;
-  unsigned char *buffer;
+  STRUCT_AUTH(auth);
+  char *server_resp_buf = NULL;
+  char *buffer;
   user_server_port_t *usp=NULL;
   protocol_t *protocol=NULL;
   token_t *our_tag=NULL;
   token_t *client_tag=NULL;
   size_t server_resp_buf_size=0;
-  flag_t tls_state=0;
+  flag_t tls_state = 0;
   timed_log_t auth_log;
-  char from_to_host_str[(NI_MAXHOST*2)+2];
+  char from_to_host_str[((NI_MAXHOST+NI_MAXSERV)*2)+2];
   char from_host_str[NI_MAXHOST];
   char to_host_str[NI_MAXHOST];
   char from_serv_str[NI_MAXSERV];
   char to_serv_str[NI_MAXSERV];
   char *servername=NULL;
-  char *username=NULL;
   char *port=NULL;
-  char *progname=NULL;
   io_t *client_io=NULL;
   io_t *server_io=NULL;
   FILE *fh;
-  int bytes_written=0;
-  int bytes_read=0;
+  size_t bytes_written = 0;
+  size_t bytes_read = 0;
   int status;
   int round_robin_server=0;
   int rnd;
   int s=-1;
   int *g = NULL;
   int rc;
+  int flag;
 
 #ifdef WITH_SSL_SUPPORT
   SSL_CTX *ssl_ctx=NULL;
 #endif /* WITH_SSL_SUPPORT */
 
-  /*
-   * Create Logger
-   */
-  vl=vanessa_logger_openlog_filehandle(stderr, LOG_IDENT, LOG_DEBUG,
-		  VANESSA_LOGGER_F_CONS|VANESSA_LOGGER_F_NO_IDENT_PID);
-  if(!vl) {
-    fprintf(stderr, "main: vanessa_logger_openlog_syslog\n"
-                    "Fatal error opening logger. Exiting.\n");
-    perdition_exit_cleanly(-1);
-  }
-  vanessa_logger_set(vl);
+  /* Create Logger */
+  progname = argv[0];
+  logger_init();
 
   /*Parse options*/
   options(argc, argv, OPT_FIRST_CALL);
 
+  /*Read config file*/
+  if(opt.config_file!=NULL){
+    config_file_to_opt(opt.config_file);
+  }
+
   /* Initialise setting of proctitle */
   init_set_proc_title(argc, argv, envp);
   progname = strdup(get_progname(argv[0]));
@@ -291,19 +384,12 @@ int main (int argc, char **argv, char **envp){
 	  VANESSA_LOGGER_ERR("Error initialising process title\n");
 	  perdition_exit_cleanly(-1);
   }
-  set_proc_title(progname);
+  set_proc_title("%s: daemon", progname);
 
-  /*
-   * Update Logger
-   */
-  if(!opt.debug){
-    vanessa_logger_change_max_priority(vl, opt.quiet?LOG_ERR:LOG_INFO);
-  }
-
-  /*Read config file*/
-  if(opt.config_file!=NULL){
-    config_file_to_opt(opt.config_file);
-  }
+  /* Update Logger */
+  if (!opt.debug)
+    vanessa_logger_change_max_priority(vanessa_logger_get(),
+				       opt.quiet?LOG_ERR:LOG_INFO);
 
   /*Open the dbserver_get library, if we have a library*/
   if(getserver_openlib(opt.map_library, opt.map_library_opt,
@@ -349,41 +435,11 @@ int main (int argc, char **argv, char **envp){
     }
   }
 
-  /*Close file descriptors and detach process from shell as necessary*/
-  if(opt.inetd_mode || opt.no_daemon || fh != NULL){
-    vanessa_socket_daemon_inetd_process();
-  }
-  else{
-    vanessa_socket_daemon_process();
-  }
-
   /*
    * Re-create logger now process is detached (unless in inetd mode)
    * and configuration file has been read.
    */
-  if(fh != NULL) {
-    vl=vanessa_logger_openlog_filehandle(fh, LOG_IDENT,
-      opt.debug?LOG_DEBUG:(opt.quiet?LOG_ERR:LOG_INFO),
-      VANESSA_LOGGER_F_CONS|VANESSA_LOGGER_F_TIMESTAMP);
-  }
-  else if(opt.log_facility!=NULL && *(opt.log_facility)=='/'){
-    vl=vanessa_logger_openlog_filename(opt.log_facility, LOG_IDENT,
-      opt.debug?LOG_DEBUG:(opt.quiet?LOG_ERR:LOG_INFO), 
-      VANESSA_LOGGER_F_CONS|VANESSA_LOGGER_F_TIMESTAMP);
-  }
-  else {
-    vl=vanessa_logger_openlog_syslog_byname(opt.log_facility, LOG_IDENT,
-      opt.debug?LOG_DEBUG:(opt.quiet?LOG_ERR:LOG_INFO), LOG_CONS);
-  }
-  if(vl == NULL) {
-	  VANESSA_LOGGER_DEBUG("vanessa_logger_openlog");
-	  VANESSA_LOGGER_ERR("Fatal error opening logger. Exiting.");
-	  perdition_exit_cleanly(-1);
-  }
-  vanessa_logger_closelog(vanessa_logger_get());
-  vanessa_logger_set(vl);
- 
-  /* Create a PID file */
+  logger_reopen(fh);
 
   /*Seed the uname structure*/
   if((system_uname=(struct utsname *)malloc(sizeof(struct utsname)))==NULL){
@@ -433,37 +489,43 @@ int main (int argc, char **argv, char **envp){
 	  our_tag = NULL;
   }
 
+  /*
+   * Log the options we will be running with.
+   * If we are in inetd mode then only do this if debugging is turned on,
+   * else debugging is a bit too verbose.
+   */
+  if ((!opt.quiet && !opt.inetd_mode && !opt.no_daemon && !fh) || opt.debug) {
+    if (log_options()) {
+      VANESSA_LOGGER_DEBUG("log_options");
+      VANESSA_LOGGER_ERR("Fatal error logging options. Exiting.");
+      perdition_exit_cleanly(-1);
+    }
+  }
+
 #ifdef WITH_SSL_SUPPORT
   /*Set up the ssl mode */
   opt.ssl_mode=protocol->encryption(opt.ssl_mode);
-  opt.capability = protocol->capability(opt.capability, 
-		  &(opt.mangled_capability), opt.ssl_mode, tls_state);
 
   if(opt.ssl_mode & SSL_LISTEN_MASK) {
     ssl_ctx = perdition_ssl_ctx(opt.ssl_ca_file, opt.ssl_ca_path,
 				opt.ssl_cert_file, opt.ssl_key_file,
-				opt.ssl_ca_chain_file, opt.ssl_listen_ciphers);
+				opt.ssl_ca_chain_file, opt.ssl_listen_ciphers,
+				PERDITION_SSL_SERVER);
     if(!ssl_ctx) {
       PERDITION_DEBUG_SSL_ERR("perdition_ssl_ctx");
       VANESSA_LOGGER_ERR("Fatal error establishing SSL context for listening");
       perdition_exit_cleanly(-1);
     }
   }
-#else
-  opt.capability = protocol->capability(opt.capability, 
-		  &(opt.mangled_capability), SSL_MODE_NONE, SSL_MODE_NONE);
 #endif /* WITH_SSL_SUPPORT */
 
-
-  /*
-   * Log the options we will be running with.
-   * If we are in inetd mode then only do this if debugging is turned on,
-   * else debugging is a bit too verbose.
-   */
-  if((!opt.quiet && !opt.inetd_mode && !opt.no_daemon && !fh) || opt.debug){
-    if(log_options()){
-      VANESSA_LOGGER_DEBUG("log_options");
-      VANESSA_LOGGER_ERR("Fatal error logging options. Exiting.");
+  /* Close file descriptors and detach process from shell as necessary */
+  if (opt.inetd_mode || opt.no_daemon || fh != NULL)
+    vanessa_socket_daemon_inetd_process();
+  else {
+    vanessa_socket_daemon_process();
+    if (vanessa_logger_reopen(vanessa_logger_get()) < 0) {
+      fprintf(stderr, "Fatal error reopening logger. Exiting.");
       perdition_exit_cleanly(-1);
     }
   }
@@ -482,9 +544,8 @@ int main (int argc, char **argv, char **envp){
    * If we are using the server's ok line then allocate a buffer to store it
    */ 
   if(opt.server_resp_line){
-    if((server_resp_buf=(unsigned char *)malloc(
-      sizeof(unsigned char)*MAX_LINE_LENGTH
-    ))==NULL){
+    server_resp_buf = malloc(MAX_LINE_LENGTH);
+    if (!server_resp_buf) {
       VANESSA_LOGGER_DEBUG_ERRNO("malloc server_resp_buf");
       VANESSA_LOGGER_ERR("Fatal error allocating memory. Exiting.");
       perdition_exit_cleanly(-1);
@@ -511,14 +572,14 @@ int main (int argc, char **argv, char **envp){
   /* Open incoming socket as required */
   if(!opt.inetd_mode) {
 	  size_t nfrom;
-	  char **fromv;
+	  const char **fromv;
 	  
   	  if (opt.bind_address) 
 		  nfrom = vanessa_dynamic_array_get_count(opt.bind_address);
 	  else 
 		  nfrom = 1;
 	  
-  	  fromv = (char **)malloc(((nfrom * 2) + 1) * sizeof(char *));
+	  fromv = malloc(sizeof(*fromv) * ((nfrom * 2) + 1));
 	  if (!fromv) {
 		  VANESSA_LOGGER_DEBUG_ERRNO("malloc fromv");
 		  VANESSA_LOGGER_ERR("Fatal error allocating memory. Exiting.");
@@ -543,8 +604,12 @@ int main (int argc, char **argv, char **envp){
 	  	  fromv[1] = opt.listen_port;
 	  }
 	  fromv[nfrom * 2] = NULL;
-	  g = vanessa_socket_server_bindv(fromv, 
-	  			opt.no_lookup?VANESSA_SOCKET_NO_LOOKUP:0);
+	  flag = 0;
+    	  if (opt.no_lookup)
+	    	flag |= VANESSA_SOCKET_NO_LOOKUP;
+    	  if (opt.tcp_keepalive)
+	    	flag |= VANESSA_SOCKET_TCP_KEEPALIVE;
+	  g = vanessa_socket_server_bindv(fromv, flag);
 	  free(fromv);
 	  if(!g) {
 		  VANESSA_LOGGER_DEBUG("vanessa_socket_server_bindv");
@@ -593,9 +658,12 @@ int main (int argc, char **argv, char **envp){
     }
   }
   else{
+    flag = 0;
+    if (opt.tcp_keepalive)
+      flag |= VANESSA_SOCKET_TCP_KEEPALIVE;
     s = vanessa_socket_server_acceptv(g, opt.connection_limit,
 				      (struct sockaddr *) peername,
-				      (struct sockaddr *) sockname, 0);
+				      (struct sockaddr *) sockname, flag);
     if(s < 0){
       VANESSA_LOGGER_DEBUG("vanessa_socket_server_accept");
       VANESSA_LOGGER_ERR("Fatal error accepting child connection. Exiting.");
@@ -652,14 +720,18 @@ int main (int argc, char **argv, char **envp){
   }
   else {
     *to_host_str='\0';
-    *from_host_str='\0';
+    *to_serv_str='\0';
   }
   *from_to_host_str='\0';
   if(peername!=NULL && sockname!=NULL){
     strcat(from_to_host_str, " ");
     strcat(from_to_host_str, from_host_str);
+    strcat(from_to_host_str, ":");
+    strcat(from_to_host_str, from_serv_str);
     strcat(from_to_host_str, "->");
     strcat(from_to_host_str, to_host_str);
+    strcat(from_to_host_str, ":");
+    strcat(from_to_host_str, to_serv_str);
   }
 
   /*Seed rand*/
@@ -668,13 +740,13 @@ int main (int argc, char **argv, char **envp){
 
   /*Log the session and change the proctitle*/
   if(opt.inetd_mode) {
-    VANESSA_LOGGER_INFO_UNSAFE("Connect:%s inetd_pid=%d",
+    VANESSA_LOGGER_INFO_UNSAFE("Connect: %s inetd_pid=%d",
           from_to_host_str, getppid());
   }
   else {
-    VANESSA_LOGGER_INFO_UNSAFE("Connect:%s", from_to_host_str);
+    VANESSA_LOGGER_INFO_UNSAFE("Connect: %s", from_to_host_str);
   }
-  set_proc_title("%s: connect", progname);
+  set_proc_title("%s: connect (%s)", progname, from_to_host_str);
 
 #ifdef WITH_SSL_SUPPORT
   if(opt.ssl_mode & SSL_MODE_SSL_LISTEN) {
@@ -688,22 +760,21 @@ int main (int argc, char **argv, char **envp){
 #endif /* WITH_SSL_SUPPORT */
 
   /*Speak to our client*/
-  if(greeting(client_io, protocol, GREETING_ADD_NODENAME)){
+  if(protocol->greeting(client_io, GREETING_ADD_NODENAME)){
     VANESSA_LOGGER_DEBUG("greeting");
-    VANESSA_LOGGER_ERR_UNSAFE("Fatal error writing to client%s."
+    VANESSA_LOGGER_ERR_UNSAFE("Fatal error writing to client %s."
 			      "Exiting child.", from_to_host_str);
     perdition_exit_cleanly(-1);
   }
 
-  pw.pw_name=NULL;
-  pw.pw_passwd=NULL;
   /* Authenticate the user*/
   for(;;){
     /*Read the USER and PASS lines from the client */
-    status=(*(protocol->in_get_pw))(client_io, &pw, &client_tag);
+    status = (*(protocol->in_get_auth))(client_io, opt.ssl_mode, tls_state,
+					&auth, &client_tag);
     token_flush();
     if(status<0){
-      VANESSA_LOGGER_DEBUG("protocol->in_get_pw");
+      VANESSA_LOGGER_DEBUG("protocol->in_get_auth");
       if (io_get_err(client_io) == io_err_timeout)
 	VANESSA_LOGGER_ERR_UNSAFE("Fatal Error: Timeout reading "
 				  "authentication information from "
@@ -716,8 +787,7 @@ int main (int argc, char **argv, char **envp){
       perdition_exit_cleanly(-1);
     }
     else if(status == 1){
-      VANESSA_LOGGER_ERR_UNSAFE("Closing NULL session:%s username=%s",
-				from_to_host_str, str_null_safe(pw.pw_name));
+      perdition_log_close_early(from_to_host_str, &auth);
       perdition_exit_cleanly(0);
     }
 #ifdef WITH_SSL_SUPPORT
@@ -730,8 +800,6 @@ int main (int argc, char **argv, char **envp){
         perdition_exit_cleanly(-1);
       }
       tls_state |= SSL_MODE_TLS_LISTEN;
-      opt.capability = protocol->capability(opt.capability,
-            &(opt.mangled_capability), opt.ssl_mode, tls_state);
       continue;
     }
     else if(opt.login_disabled ||
@@ -744,8 +812,10 @@ int main (int argc, char **argv, char **envp){
     }
 #endif /* WITH_SSL_SUPPORT */
 
-    username = username_mangle(pw.pw_name, STATE_GET_SERVER);
-    if (!username) {
+    {
+    char *id;
+    id = username_mangle(auth_get_authorisation_id(&auth), STATE_GET_SERVER);
+    if (!id) {
       VANESSA_LOGGER_DEBUG("username_mangle STATE_GET_SERVER");
       VANESSA_LOGGER_ERR_UNSAFE("Fatal error manipulating username "
 				"for client \"%s\": Exiting child",
@@ -755,25 +825,29 @@ int main (int argc, char **argv, char **envp){
 
     /*Read the server from the map, if we have a map*/
     if(dbserver_get || dbserver_get2 || opt.client_server_specification) {
-	usp = getserver(username, from_host_str, to_host_str,
+	usp = getserver(id, from_host_str, to_host_str,
 			from_serv_str, to_serv_str,
 			dbserver_get, dbserver_get2);
     }
+
+    if (id != auth_get_authorisation_id(&auth))
+      free(id);
+    }
+
     if(usp){
       port = usp->port;
       servername = usp->server;
     
       if(opt.username_from_database && usp->user){
-        if (pw.pw_name != username && pw.pw_name != NULL) {
-          free(pw.pw_name);
-          pw.pw_name = NULL;
-        }
-        if((pw.pw_name=strdup(usp->user))==NULL){
+      	char * new_id = strdup(usp->user);
+        if (!new_id) {
 	    VANESSA_LOGGER_DEBUG_ERRNO("strdup");
             VANESSA_LOGGER_ERR_UNSAFE("Fatal error manipulating username "
 				      "for client \"%s\": Exiting child",
 				      from_host_str);
         }
+        free(auth_get_authorisation_id(&auth));
+	auth_set_authorisation_id(&auth, new_id);
       }
     }
 
@@ -801,17 +875,23 @@ int main (int argc, char **argv, char **envp){
 
 #ifdef WITH_PAM_SUPPORT
     if(opt.authenticate_in){
-      pw2.pw_name = username_mangle(pw.pw_name, STATE_LOCAL_AUTH);
-      if (!pw2.pw_name) {
+      struct auth auth2 = auth;
+      char *id;
+
+      id = username_mangle(auth_get_authorisation_id(&auth), STATE_LOCAL_AUTH);
+      if (!id) {
         VANESSA_LOGGER_DEBUG("username_mangle STATE_LOCAL_AUTH");
         VANESSA_LOGGER_ERR_UNSAFE("Fatal error manipulating username for "
 				  "client \"%s\": Exiting child",
 				  from_host_str);
         perdition_exit_cleanly(-1);
       }
-      pw2.pw_passwd=pw.pw_passwd;
+      auth_set_authorisation_id(&auth2, id);
 
-      if((status=protocol->in_authenticate(&pw2, client_io, client_tag))==0){
+      status = protocol->in_authenticate(&auth2, client_io, client_tag);
+      if (id != auth_get_authorisation_id(&auth))
+      	free(id);
+      if (!status) {
         VANESSA_LOGGER_DEBUG("protocol->in_authenticate");
         VANESSA_LOGGER_INFO(
 	  "Local authentication failure for client: Allowing retry."
@@ -837,9 +917,12 @@ int main (int argc, char **argv, char **envp){
 #endif /* WITH_PAM_SUPPORT */
 
     /* Talk to the real pop server for the client*/
-    s = vanessa_socket_client_src_open(NULL, NULL, servername, port,
-				       opt.no_lookup ? 
-				       VANESSA_SOCKET_NO_LOOKUP : 0);
+    flag = 0;
+    if (opt.no_lookup)
+      flag |= VANESSA_SOCKET_NO_LOOKUP;
+    if (opt.tcp_keepalive)
+      flag |= VANESSA_SOCKET_TCP_KEEPALIVE;
+    s = vanessa_socket_client_src_open(NULL, NULL, servername, port, flag);
     if(s < 0) {
 	    VANESSA_LOGGER_DEBUG("vanessa_socket_client_open");
 	    LOGIN_FAILED_PROTOCOL(PROTOCOL_ERR, "Could not connect to server");
@@ -866,22 +949,29 @@ int main (int argc, char **argv, char **envp){
     }
 #endif /* WITH_SSL_SUPPORT */
 
+    {
+    struct auth auth2 = auth;
+    char *id;
+    int sasl_mech = 0;
+
     /* Authenticate the user with the pop server */
-    pw2.pw_name = username_mangle(pw.pw_name, STATE_REMOTE_LOGIN);
-    if(!pw2.pw_name) {
+    id = username_mangle(auth_get_authorisation_id(&auth), STATE_REMOTE_LOGIN);
+    if (!id) {
       VANESSA_LOGGER_DEBUG("username_mangle STATE_REMOTE_LOGIN");
       VANESSA_LOGGER_ERR_UNSAFE("Fatal error manipulating username "
 				"for client \"%s\": Exiting child",
 				from_host_str);
       perdition_exit_cleanly(-1);
     }
-    pw2.pw_passwd=pw.pw_passwd;
+    auth_set_authorisation_id(&auth2, id);
 
-    status = (*(protocol->out_setup))(server_io, client_io, &pw2, our_tag, 
-		    protocol);
+    status = (*(protocol->out_setup))(server_io, client_io, &auth2, our_tag);
+    sasl_mech = status & PROTOCOL_S_SASL_MASK;
     if(status==0){
 	    quit(server_io, protocol, our_tag);
 	    LOGIN_FAILED_PROTOCOL(PROTOCOL_NO, "Connection Negotiation Failure");
+	    if (id != auth_get_authorisation_id(&auth))
+	      free(id);
 	    PERDITION_CLEAN_UP_MAIN;
 	    continue;
     }
@@ -890,6 +980,7 @@ int main (int argc, char **argv, char **envp){
       VANESSA_LOGGER_ERR("Fatal error negotiating setup. Exiting child.");
       perdition_exit_cleanly(-1);
     }
+
 #ifdef WITH_SSL_SUPPORT
     else if((opt.ssl_mode & SSL_MODE_TLS_OUTGOING) &&
           (status & PROTOCOL_S_STARTTLS)) {
@@ -907,6 +998,8 @@ int main (int argc, char **argv, char **envp){
 		    !(status & PROTOCOL_S_STARTTLS)) {
 	    quit(server_io, protocol, our_tag);
 	    LOGIN_FAILED_PROTOCOL(PROTOCOL_NO, "TLS not present");
+	    if (id != auth_get_authorisation_id(&auth))
+	        free(id);
 	    PERDITION_CLEAN_UP_MAIN;
 	    continue;
     }
@@ -916,16 +1009,20 @@ int main (int argc, char **argv, char **envp){
       server_resp_buf_size=MAX_LINE_LENGTH-1;
     }
     token_flush();
-    status = (*(protocol->out_authenticate))(server_io, client_io, &pw2, 
-		    our_tag, protocol, server_resp_buf, &server_resp_buf_size);
+    status = (*(protocol->out_authenticate))(server_io, client_io, tls_state,
+					     &auth2, sasl_mech, our_tag,
+					     protocol, server_resp_buf,
+					     &server_resp_buf_size);
+    if (id != auth_get_authorisation_id(&auth))
+        free(id);
     if(status==0) {
 	    quit(server_io, protocol, our_tag);
             if(opt.server_resp_line){
               sleep(PERDITION_AUTH_FAIL_SLEEP);
               *(server_resp_buf+server_resp_buf_size)='\0';
-              if(protocol->write(client_io, WRITE_STR_NO_CLLF, client_tag, 
-                    NULL, 1, "%s", server_resp_buf)<0){
-                VANESSA_LOGGER_DEBUG("protocol->write");
+              if (protocol->write_str(client_io, WRITE_STR_NO_CLLF, client_tag,
+				      NULL, server_resp_buf) < 0) {
+                VANESSA_LOGGER_DEBUG("protocol->write_str quit");
                 VANESSA_LOGGER_ERR("Fatal error writing to client. Exiting child.");
                 perdition_exit_cleanly(-1);
               }
@@ -942,6 +1039,12 @@ int main (int argc, char **argv, char **envp){
 	    PERDITION_CLEAN_UP_MAIN;
 	    continue;
     }
+    if (status==3) {
+	    sleep(VANESSA_LOGGER_ERR_SLEEP);
+	    LOGIN_FAILED_PROTOCOL(PROTOCOL_NO, "SASL mechanism unavailable");
+	    PERDITION_CLEAN_UP_MAIN;
+	    continue;
+    }
     else if(status<0){
       VANESSA_LOGGER_DEBUG_UNSAFE("protocol->out_authenticate %d", status);
       if (io_get_err(client_io) == io_err_timeout)
@@ -954,21 +1057,22 @@ int main (int argc, char **argv, char **envp){
 
     if(opt.server_resp_line){
       *(server_resp_buf+server_resp_buf_size)='\0';
-      if(protocol->write(client_io, WRITE_STR_NO_CLLF, client_tag, 
-            NULL, 1, "%s", server_resp_buf)<0){
-        VANESSA_LOGGER_DEBUG("protocol->write");
+      if (protocol->write_str(client_io, WRITE_STR_NO_CLLF, client_tag,
+			      NULL, server_resp_buf) < 0) {
+        VANESSA_LOGGER_DEBUG("protocol->write_str logged in, server_resp");
         VANESSA_LOGGER_ERR("Fatal error writing to client. Exiting child.");
         perdition_exit_cleanly(-1);
       }
     }
     else{
-      if(protocol->write(client_io, NULL_FLAG, client_tag, 
-            protocol->type[PROTOCOL_OK], 0, opt.ok_line)<0){
-        VANESSA_LOGGER_DEBUG("protocol->write");
+      if (protocol->write_str(client_io, NULL_FLAG, client_tag,
+			      protocol->type[PROTOCOL_OK], opt.ok_line) < 0) {
+        VANESSA_LOGGER_DEBUG("protocol->write_str logged in");
         VANESSA_LOGGER_ERR("Fatal error writing to client. Exiting child.");
         perdition_exit_cleanly(-1);
       }
     }
+    }
 
     break;
   }
@@ -980,7 +1084,8 @@ int main (int argc, char **argv, char **envp){
   }
 
   /*We need a buffer for reads and writes to the server*/
-  if((buffer=(unsigned char *)malloc(BUFFER_SIZE*sizeof(unsigned char)))==NULL){
+  buffer = malloc(BUFFER_SIZE);
+  if (!buffer) {
     VANESSA_LOGGER_DEBUG_ERRNO("malloc");
     VANESSA_LOGGER_ERR("Fatal error allocating memory. Exiting child.");
     perdition_exit_cleanly(-1);
@@ -991,21 +1096,24 @@ int main (int argc, char **argv, char **envp){
   io_set_timeout(server_io, opt.timeout);
   if(io_pipe(server_io, client_io, buffer, BUFFER_SIZE,
         &bytes_written, &bytes_read, &auth_log)<0){
-    VANESSA_LOGGER_DEBUG("vanessa_socket_pipe");
-    if (io_get_err(client_io) == io_err_timeout)
-      VANESSA_LOGGER_ERR("Fatal error: Timeout piping data. "
-			 "Exiting child.");
-    else
+    if (io_get_err(client_io) == io_err_timeout) {
+      VANESSA_LOGGER_INFO("Timeout piping data.");
+      if (protocol->bye && protocol->bye(client_io, "Timeout") < 0) {
+        VANESSA_LOGGER_DEBUG("protocol->bye timeout");
+        VANESSA_LOGGER_ERR("Fatal error writing to client. Exiting child.");
+        perdition_exit_cleanly(-1);
+      }
+    } else {
+      VANESSA_LOGGER_DEBUG("vanessa_socket_pipe");
       VANESSA_LOGGER_ERR("Fatal error piping data. Exiting child.");
-    perdition_exit_cleanly(-1);
+      perdition_exit_cleanly(-1);
+    }
   }
 
   /*Time to leave*/
-  VANESSA_LOGGER_INFO_UNSAFE("Close:%s user=\"%s\" received=%d sent=%d",
-			     str_null_safe(from_to_host_str),
-			     str_null_safe(pw.pw_name),
-			     bytes_read, bytes_written);
-  set_proc_title("%s: close", progname);
+  perdition_log_close(from_to_host_str, &auth, bytes_read, bytes_written);
+  set_proc_title("%s: close (%s)", progname,
+		 str_null_safe(auth_get_authorisation_id(&auth)));
 
   getserver_closelib(handle);
   perdition_exit_cleanly(0);
@@ -1201,8 +1309,8 @@ write_pid_file(const char *pidfilename, const char *username,
 
 	int     pidfilefd;
 	char    pidbuf[11];
-	pid_t   pid;
-	size_t  bytes;
+	unsigned long	pid;
+	ssize_t  bytes;
 
 	if (create_pid_directory(pidfilename, username, group) < 0) {
 		return -1;
@@ -1266,6 +1374,13 @@ write_pid_file(const char *pidfilename, const char *username,
 			return -1;
 		}
 
+		if ((long)pid != (pid_t)pid) {
+			VANESSA_LOGGER_DEBUG_UNSAFE("Invalid pid in pid-file "
+	 				"[%s]: larger than maximum pid",
+					pidfilename);
+			return -1;
+		}
+
 		if (!kill(pid, 0)) {
 			VANESSA_LOGGER_ERR_UNSAFE("Fatal error: "
 					"Pid file [%s] exists for "
@@ -1301,7 +1416,7 @@ write_pid_file(const char *pidfilename, const char *username,
 	}
 
 	if (snprintf(pidbuf, sizeof(pidbuf), "%u", 
-				getpid()) >= sizeof(pidbuf)) {
+				getpid()) >= (int)sizeof(pidbuf)) {
 		VANESSA_LOGGER_DEBUG_UNSAFE("Pid too long for buffer [%u]", 
 				getpid());
 		goto unlink;
@@ -1309,7 +1424,7 @@ write_pid_file(const char *pidfilename, const char *username,
 
 	while (1) {
 		bytes = write(pidfilefd, pidbuf, strlen(pidbuf));
-		if (bytes != strlen(pidbuf)) {
+		if (bytes != (ssize_t)strlen(pidbuf)) {
 			if (bytes < 0 && errno == EINTR) {
 				continue;
 			}
